import os
import pickle
from django.apps import AppConfig
from .dga_classifier_scripts.n_grams import N_grams
from .dga_classifier_scripts.known_tld import KnownTLD
from .dga_classifier_scripts.multiclass_dga_clf import MulticlassClassifier


def load_xgboost_model(model_name="None"):
    """Function loads model according to given filename

    Args:
        model_name (str, optional): Name of the file, 
                                    where the model is stored. 
                                    Defaults to "None".

    Returns:
        _type_: XGBoost model
    """
    if model_name == "None":
        print("Please enter the name of file of the model.")
        return None
    
    # Get the absolute path to the 'models' directory within the app
    app_path = os.path.dirname(os.path.abspath(__file__))
    models_dir = os.path.join(app_path, 'dga_classifier_scripts/models')

    # Load the XGBoost model
    model_path = os.path.join(models_dir, model_name)
    with open(model_path, 'rb') as model_file:
        xgb_model = pickle.load(model_file)

    return xgb_model


def get_model_path(model_name:str=None):
    """Returns the path to the model name
    All models are saved in one directory of the Django app
    
    Args:
        model_name (str, optional): Model filename. Defaults to None.
    """
    if not model_name:
        print("Enter the model filename.")
        return None
    app_path = os.path.dirname(os.path.abspath(__file__))
    model_filepath = os.path.join(app_path, f'dga_classifier_scripts/models/{model_name}')
    return model_filepath

def get_known_tlds():
    """Function creates and intializes the KnownTLD object
    which contains all well-known TLDs. The well-known
    TLDs are used as a feature in DGA classification, because
    in could indicate benign domain.

    Returns:
        tld: Object which holds known-tlds 
    """
    app_path = os.path.dirname(os.path.abspath(__file__))
    tld_dir = os.path.join(app_path, 'dga_classifier_scripts/models/')
    tld_path = tld_dir + "public_suffix_list.dat.txt"
    tld = KnownTLD(tld_path)
    return tld

def get_ngrams():
    """Function creates and intializes the N_grams object. The n-grams
    are used as a feature in DGA classification.

    Returns:
        N_grams: Object which holds n-grams 
    """
    app_path = os.path.dirname(os.path.abspath(__file__))
    ngram_dir = os.path.join(app_path, 'dga_classifier_scripts/models/')
    dga_ngram_csv = ngram_dir + "dga-ngram.csv"
    nondga_ngram_csv = ngram_dir + "non-dga-ngram.csv" 
    return N_grams(dga_ngram_csv=dga_ngram_csv, nondga_ngram_csv=nondga_ngram_csv)

class DgaClassifierConfig(AppConfig):
    default_auto_field = 'django.db.models.BigAutoField'
    name = 'dga_classifier'
    binary_model = load_xgboost_model(model_name="01_xgb_binary_clf.pkl")
    multiclass_model_path = get_model_path(model_name="01_dga_xgb_multiclass_clf.pkl") 
    multiclass_model = MulticlassClassifier(multiclass_model_path)
    tlds = get_known_tlds()
    ngrams = get_ngrams()
